#include "display.h"

#define DEFAULT_FIFO_SIZE	(256*1024)

u32	p_frameBufferIndex;
u32 p_isFirstFrame;
void *p_frameBuffer[2] = { NULL, NULL};
GXRModeObj *p_rMode;

void display_init() {
	f32 yscale;
	void *gp_fifo = NULL;
	u32 xfbHeight;
	Mtx44 perspective;
	GXColor background = {0, 0, 0, 0xff};

	p_isFirstFrame = 1;
	p_frameBufferIndex = 0;
	
	VIDEO_Init();
 
	p_rMode = VIDEO_GetPreferredMode(NULL);
	
	// allocate 2 framebuffers for double buffering
	p_frameBuffer[0] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(p_rMode));
	p_frameBuffer[1] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(p_rMode));

	VIDEO_Configure(p_rMode);
	VIDEO_SetNextFramebuffer(p_frameBuffer[p_frameBufferIndex]);
	VIDEO_SetBlack(FALSE);
	VIDEO_Flush();
	VIDEO_WaitVSync();
	if(p_rMode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();
	
	p_frameBufferIndex ^= 1;

	// setup the fifo and then init the flipper
	gp_fifo = memalign(32,DEFAULT_FIFO_SIZE);
	memset(gp_fifo,0,DEFAULT_FIFO_SIZE);
 
	GX_Init(gp_fifo,DEFAULT_FIFO_SIZE);
	
	// clears the bg to color and clears the z buffer
	GX_SetCopyClear(background, 0x00ffffff);
	
	// other gx setup
	GX_SetViewport(0,0,p_rMode->fbWidth,p_rMode->efbHeight,0,1);
	yscale = GX_GetYScaleFactor(p_rMode->efbHeight,p_rMode->xfbHeight);
	xfbHeight = GX_SetDispCopyYScale(yscale);
	GX_SetScissor(0,0,p_rMode->fbWidth,p_rMode->efbHeight);
	GX_SetDispCopySrc(0,0,p_rMode->fbWidth,p_rMode->efbHeight);
	GX_SetDispCopyDst(p_rMode->fbWidth,xfbHeight);
	GX_SetCopyFilter(p_rMode->aa,p_rMode->sample_pattern,GX_TRUE,p_rMode->vfilter);
	GX_SetFieldMode(p_rMode->field_rendering,((p_rMode->viHeight==2*p_rMode->xfbHeight)?GX_ENABLE:GX_DISABLE));

	if (p_rMode->aa) {
		GX_SetPixelFmt(GX_PF_RGB565_Z16, GX_ZC_LINEAR);
	} else {
		GX_SetPixelFmt(GX_PF_RGB8_Z24, GX_ZC_LINEAR);
	}
	
	GX_SetCullMode(GX_CULL_NONE);
	GX_CopyDisp(p_frameBuffer[p_frameBufferIndex],GX_TRUE);
	GX_SetDispCopyGamma(GX_GM_1_0);

	// setup the vertex descriptor
	// tells the flipper to expect direct data
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS, GX_POS_XY, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_F32, 0);
	

	GX_SetNumChans(1);
	GX_SetNumTexGens(1);
	GX_SetTevOp(GX_TEVSTAGE0, GX_REPLACE);
	GX_SetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
	GX_SetTexCoordGen(GX_TEXCOORD0, GX_TG_MTX2x4, GX_TG_TEX0, GX_IDENTITY);


	GX_InvalidateTexAll();
	
	guOrtho(perspective,0, SCREEN_HEIGHT-1, 0, SCREEN_WIDTH-1, 0, 300 );
	GX_LoadProjectionMtx(perspective, GX_ORTHOGRAPHIC);
}

void display_preFrame() {
	Mtx GXmodelView2D;
	
	GX_SetViewport(0,0,p_rMode->fbWidth,p_rMode->efbHeight,0,1);
	GX_InvVtxCache();
	GX_InvalidateTexAll();

	GX_ClearVtxDesc();
	GX_SetVtxDesc(GX_VA_POS, GX_DIRECT);
	GX_SetVtxDesc(GX_VA_TEX0, GX_DIRECT);

	guMtxIdentity(GXmodelView2D);
	guMtxTransApply (GXmodelView2D, GXmodelView2D, 0.0F, 0.0F, -5.0F);
	GX_LoadPosMtxImm(GXmodelView2D,GX_PNMTX0);
}

void display_postFrame() {
	GX_DrawDone();
		
	GX_SetZMode(GX_TRUE, GX_LEQUAL, GX_TRUE);
	GX_SetBlendMode(GX_BM_BLEND, GX_BL_SRCALPHA, GX_BL_INVSRCALPHA, GX_LO_CLEAR);
	GX_SetAlphaUpdate(GX_TRUE);
	GX_SetColorUpdate(GX_TRUE);
	GX_CopyDisp(p_frameBuffer[p_frameBufferIndex],GX_TRUE);

	VIDEO_SetNextFramebuffer(p_frameBuffer[p_frameBufferIndex]);
	if(p_isFirstFrame) {
		VIDEO_SetBlack(FALSE);
		p_isFirstFrame = 0;
	}
	VIDEO_Flush();
	VIDEO_WaitVSync();
	p_frameBufferIndex ^= 1;		// flip framebuffer
}

void display_draw( f32 x, f32 y, f32 width, f32 height,
				   f32 texLeft, f32 texTop, f32 texRight, f32 texBottom ) {

	GX_Begin(GX_QUADS, GX_VTXFMT0, 4);
		GX_Position2f32( x, y );
		GX_TexCoord2f32( texLeft, texTop );
		GX_Position2f32( x+width-1.0f, y );
		GX_TexCoord2f32( texRight, texTop  );
		GX_Position2f32( x+width-1.0f, y+height-1.0f );
		GX_TexCoord2f32( texRight, texBottom );
		GX_Position2f32( x, y+height-1.0f );
		GX_TexCoord2f32( texLeft, texBottom );
	GX_End();

}

void display_draw_full( f32 x, f32 y, f32 width, f32 height ) {
	display_draw( x, y, width, height, 
				  0.0f, 0.0f, 1.0f, 1.0f );
}